/**
 * redpen: a text inspection tool
 * Copyright (c) 2014-2015 Recruit Technologies Co., Ltd. and contributors
 * (see CONTRIBUTORS.md)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package cc.redpen.validator.sentence;

import cc.redpen.config.Configuration;
import cc.redpen.config.ValidatorConfiguration;
import cc.redpen.validator.BaseValidatorTest;
import cc.redpen.validator.ValidationError;
import org.junit.Before;
import org.junit.Test;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static org.junit.Assert.assertEquals;


public class SuggestExpressionValidatorTest extends BaseValidatorTest {
    private SuggestExpressionValidator validator;

    public SuggestExpressionValidatorTest() {
        super("SuggestExpression");
    }

    @Before
    public void init() {
        validator = new SuggestExpressionValidator();
        Map<String, String> synonymSamples = new HashMap<>();
        synonymSamples.put("like", "such as");
        synonymSamples.put("info", "information");
        synonymSamples.put("こんにちは", "良い一日");
        validator.setSynonyms(synonymSamples);
    }

    @Test
    public void testSynonym() {
        List<ValidationError> errors = new ArrayList<>();
        validator.setErrorList(errors);
        validator.validate(sentence("it like a piece of a cake."));
        assertEquals(1, errors.size());
    }

    @Test
    public void testWithoutSynonym() {
        List<ValidationError> errors = new ArrayList<>();
        validator.setErrorList(errors);
        validator.validate(sentence("it love a piece of a cake."));
        assertEquals(0, errors.size());
    }

    @Test
    public void testWithMultipleSynonyms() {
        List<ValidationError> errors = new ArrayList<>();
        validator.setErrorList(errors);
        validator.validate(sentence("it like a the info."));
        assertEquals(2, errors.size());
    }

    @Test
    public void matchWholeExpressionsOnly() {
        List<ValidationError> errors = new ArrayList<>();
        validator.setErrorList(errors);
        validator.validate(sentence("the information."));
        assertEquals(0, errors.size());
    }

    @Test
    public void japanese() {
        List<ValidationError> errors = new ArrayList<>();
        validator.setErrorList(errors);
        validator.validate(sentence("こんにちは世界"));
        assertEquals(1, errors.size());
    }

    @Test
    public void testWitoutZeroLengthSentence() {
        List<ValidationError> errors = new ArrayList<>();
        validator.setErrorList(errors);
        validator.validate(sentence(""));
        assertEquals(0, errors.size());
    }

    @Test
    public void testErrorMessageIsProperlyFormatted() {
        List<ValidationError> errors = new ArrayList<>();
        validator.setErrorList(errors);
        validator.validate(sentence("Thank you for the info."));
        assertEquals(1, errors.size());
        assertEquals("Found invalid word \"info\". Use the synonym \"information\" instead.", errors.get(0).getMessage());
    }

    @Test
    public void initDoesNotFailIfDictionaryIsNotSpecified() throws Exception {
        validator.preInit(new ValidatorConfiguration("SuggestExpression"), Configuration.builder().build());
    }
}
